/*
STABILISATION
stabilise emissions, concentration, forcing, temperature, SLR

 History:
 Stabilise concentration can generate IPCC "S"/"WRE" scenarios, initially developed Arendal autumn 2001
stabilise temperature iteration method improved Bern March 2002, extended to forcing/SLR LLN March 2003
stabilise emissions originally developed spring 2000 (based on GCI-C&C)

 */
package jcm.mod.obj;

import jcm.core.*;
import jcm.core.par.param;
import jcm.core.cur.curve;
import jcm.core.ob.module;
import jcm.mod.math.*;
import jcm.mod.carbon.carboncycle;
import jcm.mod.obj.globco2emit;
import jcm.mod.cli.*;
import jcm.mod.regemit.shares;
import jcm.mod.scen.sresBase;
import jcm.mod.scen.sresdata;
import static jcm.core.report.*;
import static jcm.gui.gen.colfont.*;
import static jcm.core.complexity.*;
import static jcm.mod.obj.stabilisation.indicators.*;
import static jcm.mod.obj.stabilisation.curvetype.*;
import static jcm.mod.obj.controller.objopt.*;
import static jcm.mod.obj.controller.butd.*;



public class stabilisation extends module implements Runnable {    //*****************************************************
    //INTERACTIONS
    public void initsetup() {
        indicator.priority=1.1;
    }
    
    public void setinteractions() {
        affectsfutureonly = true;
        //follows(history.class); //removed aug06 (distorts interacmap): shouldn't be necessary: calcstep will always run after history, only be careful re precalcs
        //setaffectedby(get(controller.class).objective); //shouldn't be necessary - as controller affects globco2emit which affects one of the target indicators
        //setaffectedby(get(shares.class), get(shares.class).kyotop.istrue());
        //setaffectedby(get(shares.class).kyotop);

        //conditional interactions for parameter help to clarify options to user, but be careful of the statement order, as these also set *not*affectedby when condition not true
        //P1 code could be neater if used an enum for the options.
        //follows(history.class, ic=="constant" || ic=="stabemit"); //this seems to be necessary if opening for the first time - doesn't help
        controller co=gm(controller.class);
        follows(shares.class, co.botuptopdo.chosen!= top_down_only); //in this case - starts after the end of the bottom-up phase of shares
        setaffectedby(co.botuptopdo);
        ic=indicator.chosen; //note repeated in precalc - runs after this
        setaffectedby( ic== constant, gm(globco2emit.class));
        setaffectedby(ic == define_emit, gm(globco2emit.class), stabemitlevel, stabemityear,stabemitendlevel,  stabemitdeclinerate, sygrowth, quartic);
        setaffectedby(ic == define_emit && quartic.istrue(), integralgt);
        setaffectedby(ic == stabconc, gm(carboncycle.class), stabconclevel, concscenmenu);
        setaffectedby(ic == stabrf_allghgaero || ic == stabrf_allghg, gm(radfor.class), stabrflevel);
        //why is it necessary to specify baseyear and baserange - is it possible that these changed but glotemp isn't flagged? 
        setaffectedby(ic == stabtemp || ic == stfuzzy, gm(glotemp.class), gm(glotemp.class).baseyear, gm(glotemp.class).baserange, stabtemplevel);
        setaffectedby(ic == stabconc && peaking.istrue(), stabconcendlevel);
        setaffectedby((ic == stabrf_allghgaero || ic == stabrf_allghg) && peaking.istrue(), stabrfend);
        setaffectedby(ic == stabtemp && peaking.istrue(), stabtempend);
        setaffectedby(ic == stabsea, gm(sealevel.class), stabsealevel, stabseayear);
        setaffectedby(ic == stabconc || ic == stabrf_allghgaero || ic == stabrf_allghg || ic == stabtemp || ic == stfuzzy, stabyear, stabconcinitslope, curveopt);
        setaffectedby(ic == stabtemp || ic == stabconc || ic == stabrf_allghgaero || ic == stabrf_allghg , peaking);
        setaffectedby(co.botuptopdo.chosen== top_down_only, wre); // if shares works bottom up, start is already delayed and defined by shares
        setaffectedby(wre, ic != constant); //always, as default, just here for clarity
        setaffectedby(wre.istrue(), stabconcstartyear);

        setaffectedby(ic == stfuzzy, gm(stabtempfuzzy.class), damp);
        setaffects(ic == stfuzzy, gm(stabtempfuzzy.class));

        //P2 add arrow association for stabemit here!
        gm(carboncycle.class).conc.associate(stabconclevel);
        gm(radfor.class).allrf.associate(stabrflevel);
        gm(radfor.class).allrf2.associate(stabrflevel);
        gm(glotemp.class).temp.associate(stabtemplevel);
        gm(sealevel.class).sealevel.associate(stabsealevel);
    }    
    
    //
    //*************** ADJUSTABLE PARAMETERS ***********************************
    //
    //
    public enum indicators { define_emit, stabconc, stabrf_allghg, stabrf_allghgaero, stabtemp, stabsea, constant, stfuzzy};
    indicators ic; //shortcut to currently chosen
    public enum curvetype { padequartic, padequintic };
    
    //
    //main menu - to choose which indicator to stabilise
    complexity[] indiccomplex = new complexity[]{normal, simplest, expert, normal, normal, experimental, expert, experimental};
    public param<indicators> indicator = new param("indicator", indicators.values(), stabconc, indiccomplex) {
        public void precalc() {
            resetguess();
        }
    };    
    //
    //IPCC-stabilisation-scenario menu
    int[] concscenlevel = {400, 450, 500, 550, 600, 650, 750, 850, 1000};
    String[] concscenname = {"400ppm", "450ppm", "500ppm", "550ppm", "600ppm", "650ppm", "750ppm", "850ppm", "1000ppm", "user"};
    public param concscenmenu = new param("stabconcmenu", concscenname, "user") {
        public void precalc() {
            if (chosen != "user") {
                stabconclevel.putval(concscenlevel[getchosenindex()]);
                stabyear.putval(2100 + (concscenlevel[getchosenindex()] - 450) / 2.0); //set endyear as a func of endlevel as per IPCC S/WRE
            }
        }
    };
    
    public param //for stabilise concentration, forcing, temperature
            stabyear = new param("stabyear", "year", 2100, 2030, (gey) - 20, grey, simplest), //stabilisation year
             stabconclevel = new param("stabconc&atlevel", "ppm", 450, 300, 1000, stabyear, black, simplest), //final concentration in stab year
             stabconcinitslope = new param("stabconcinitslope", "ppm&per&year&squared", 0, -0.15, 0.15, expert, black), //
             stabconcendlevel = new param("stabconc&endlevel", "ppm", 475, 300, 1000, stabyear, black, expert), //final concentration in gey
             stabrflevel = new param("stabrf&atlevel", "w&per&m2", 3.1, 1.5, 8.0, stabyear, green), //final total rf
             stabtemplevel = new param("stabtemp&atlevel", "degcbase", 2.0, 1.0, 7.0, stabyear, brown), //final temperature in stab year
             stabtempend = new param("temp_end&level", "degcbase", 2.0, 1.0, 7.0, stabyear, brown), //final temperature in gey
             stabrfend=new param("stabrf_end&level", "w&per&m2", 3.1, 1.5, 8.0, stabyear, dkgreen), //final total rf
             peaking = new param("peaking", false, expert), // come down after peak (to end level)
             wre = new param("wreopt", false, expert), //delay start year as Wigley-Richels-Edmonds proposal
             stabconcstartyear = new param("stabconcstartyear", "", 2020, 1990, 2050, expert); //startyear fixed by WRE
            //for stabilise emissions
     public param stabemityear = new param("stabemityear", "year", 2080, 2010, (gey) - 20, red), //stabilisation year
             stabemitlevel = new param("stabemit&atlevel", "mega&ton&carbon&per&yr", 2000, 0, 30000, stabemityear, red), //emissions in target year
             stabemitendlevel = new param("stabemit&minlevel", "mega&ton&carbon&per&yr", 500, -2000, 12000, dkred, expert), //final minimum emissions in gey
             stabemitdeclinerate = new param("stabemit&declinerate", "percent&per&yr", 2, 0, 5, brown, expert), //decline rate %/yr
             sygrowth = new param("initgrow", "percent&per&yr", 1.5, -3.5, 3.5, expert, orange), //initial growth rate
             quartic = new param("integralopt", false, expert), //control integral (quartic curve)
             integralgt = new param("integralgt", "giga&ton&carbon", 500, 150, 6000, expert); //total emissions 2000-endyear in GtC
            //for stabilise get(sealevel.class)
      public param   stabseayear = new param("stabseayear", "year", 2250, 2030, (gey) - 20, expert), //stabilisation year
             stabsealevel = new param("stabsea&atlevel", "metres", 0.75, 0.5, 3.0, stabseayear, expert), //final total rf
             damp = new param("dampopt", false, expert); 						//use damping in fuzzy control
       public param<curvetype>   curveopt = new param("curveopt", curvetype.values(), curvetype.padequintic, expert);
    //         
    //*******************MAIN LOOP CALCULATIONS **********************************
    //
    public void precalc() {
        ic=indicator.chosen;
        if (!running) { setstartyear(); } // deb("stabilisation start year="+sy);}
    }

    public void calcstep() {
        carboncycle carboncycle = gm(carboncycle.class);
        /*
        if start year,  set target stabilisation curves ( must be within calcstep loop as kyoto may affect start)
        note: for stabtemp sy >= get(glotemp.class).baseyear+basrange, to get baseyear offset correction right (done in climate module at baseyear/baserange)
         */
        if (year == sy) {
            if (ic == define_emit) {
                emittarget((int) stabemityear.getval(),  stabemitlevel.getval(), stabemitendlevel.getval(), stabemitdeclinerate.getval());
            }
            if (ic == stabconc) {
                if (peaking.istrue()) conctarget((int) stabyear.getval(), stabconclevel.getval(), stabconcendlevel.getval());
                else conctarget((int) stabyear.getval(), stabconclevel.getval());
            }
            //what was this for? if (!get(carboncycle.class).histdeforbymassbal.istrue()) enclosing all conctarget method

            if ((ic == stabtemp || ic == stabrf_allghgaero || ic == stabrf_allghg || ic == stabsea)) {
                setlevel();  //do this even within iteration, in case parameters continuously adjusted by user
                if (!running) {
                    initialguess();
                    new Thread(this, "Stabilisation").start();
                }
                conctarget(scy, co2stab, co2stabslope, co2end); //make new target co2 curve
            }
        }

        if (year >= sy) {

            if (ic == define_emit) carboncycle.totemit.set(year, target[year - gsy]);
            else if (ic.name().startsWith("stab")) carboncycle.totemit.set(year, carboncycle.inverseco2(target[year - gsy]));
            //(year-sy)<10 ? 5 : 1));
            //try to remove initial oscillations (due to feedbacks) by averaging with emissions trend - doesn't help - just increases wavelength
            //if ((year-sy)<=10)  get(carboncycle.class).totemit.set(year, (get(carboncycle.class).totemit.get(year)+2f*get(carboncycle.class).totemit.get(year-1) - get(carboncycle.class).totemit.get(year-2))/2f);

            if (ic == constant) carboncycle.totemit.set(year, carboncycle.totemit.get(year - 1));
            if (ic == stfuzzy) gm(stabtempfuzzy.class).calcstep();
        }

    } //end calcstep

    public void postcalc() {
        //correct target curve, and check whether goodenough
        if (ic == stabrf_allghgaero || ic == stabrf_allghg || (ic == stabtemp) || ic == stabsea) applycorrection();
    }    //*****************************************************
    //target curve
    public float[] target = new float[gey - gsy + 2];

    public void changeendyear(int oldey, int newey) {
        target = new float[newey - gsy + 2];
    }
    curve curvetocheck;    //variables relating to initial gradient
    public int x0,  sy = fsyfos,  scy;
    float y0, dy0, d2y0;    //iteration variables to be remembered between calls
    double targetlevel = 0, endtargetlevel = 0;
    double co2stab = 0, co2staborig = 0, stabcor = 1.0, oldstabcor = 1.0;
    double co2stabslope = 0, co2stabslopeorig = 0, stabslopecor = 1.0, oldstabslopecor = 0;
    //double co2mid=0, co2midorig=0, midcor=1.0, oldmidcor=1.0;
    double co2end = 0, co2endorig = 0, endcor = 1.0, oldendcor = 1.0;
    //	double co2endslope=0, co2endslopeorig=0, endslopecor=1.0, oldendslopecor=0;
    double olditw = Float.MAX_VALUE, oldco2stab;
    public boolean goodenough = false,  running = false;
    public int stit = 0; //number of iterations

    //**************************
    //INITIAL YEAR/GRADIENT
    void setstartyear() {
        controller co=gm(controller.class);
        //added post-copenhagen 2010, option to start bottom-up after pledges
        if (co.botuptopdo.chosen== start_bottom_up) { sy=(int)co.startyear_topdown.getval(); return; }
        if (co.botuptopdo.chosen== bottom_up_only) { sy=999999999; return; } 
        //P1 CHECK land use in futbasescen now uses fsy instead of sy, this might cause problem in WRE/kyoto variants
        //set startyear for stabilisation curve, 2000 or after kyoto or after WRE delay
        sy = fsyfos; //(gm(shares.class).kyotop.istrue() ? 2013 : fsyfos);
        if (wre.istrue()) delaystart(); //&& !gm(shares.class).kyotop.istrue()) delaystart();
    }
    //work out initial gradient and d2y/dx2
    float[] gradient(curve q, int x0) {
        float y = q.get(x0);
        /* if start before fsyfos+10 fix d2y0 to avoid effect of kinks in historical data
        note this will only work for q=carboncycle.atppm
        dy shouldn't be fixed because makes step in emissions if change carboncycle parameters, but we need to smooth it - use 5 years to avoid Pinatubo effect.
         */
        float d2ycalc = (y - 2f * q.get(x0 - 1) + q.get(x0 - 2)); /*0.02f 0.016f*/
        float d2y = (x0 < fsyfos + 10) ? (float) stabconcinitslope.getval() : d2ycalc;
        float dy = 3.5f * d2ycalc + (q.get(x0) - q.get(x0 - 5)) / 5f;
        return new float[]{y, dy, d2y};
    }
    //******** DELAY START *************************
    void delaystart() {
        /* old wre opt - WRE scenarios no longer compatible with shifting future start to 2012, 
        //if (wre.changed) stabconcstartyear.putval(2009 + (int) ((stabconclevel.getval() - 400.0) / 20.0));
        //note: used to be 2003, 350, 23 to match original WRE then updated to 2007, 400, 20, now 2009+
        */
        delaystart(sy, (int) stabconcstartyear.getval());
        sy = (int) stabconcstartyear.getval();
    }
//beware that globco2emit comes after this in loop order
    void delaystart(int sy, int dsy) {
        param scenario = gm(sresBase.class).scenario;
        int oldsc = scenario.getchosenindex();
        scenario.choose(6);
        for (int y = sy; y <= dsy; y++) gm(carboncycle.class).fossil.set(gm(globco2emit.class).fossilbase.get(y));
        scenario.choose(oldsc);
    }
//	Old note?:  WRE started by following IS92a *(6.7/7.1) to connect with 2000 historical data
    //***************************************************
    //STABILISE EMISSIONS
    //changed jan09 to permit open model directly into stabemit
    void emittarget(int xs, double ys) {
        emittarget(xs, gey, ys, ys);
    }

    void emittarget(int xs, double ys, double ye, double decrate) {
        double dy0 = y0 * sygrowth.getval() / 100f, y0 = gm(carboncycle.class).totemit.get(sy - 1) + dy0;
        double dys =  -(ys-ye)*decrate/100f ;
        if (quartic.istrue()) mathcurve.quarticend(target, sy, xs, y0, dy0, ys, dys, integralgt.getval() * 1000f);
        else mathcurve.cubicend(target, sy, xs, y0, dy0, ys, dys);
        for (int x = xs+1; x <= gey; x++)  target [x - gsy] = (float) (ye + (target [x - gsy -1]-ye) * (100f- decrate) / 100f);
    }

    /*
     *  try doing it all in one cubic doen't work - need a quartic with another constraint
    void emittarget(int xs, double ys, double ye) {
    double y0 = gm(carboncycle.class).totemit.get(sy), dy0 = y0 * sygrowth.getval() / 100f;
    mathcurve.cubictwo(target, sy, xs, gey, y0, dy0, ys, ye);
    }
     */
    //**************************
    //SET CONCN TARGET
    void conctarget(int xs, double ys) {
        conctarget(xs, ys, 0, ys);
    }

    void conctarget(int xs, double ys, double ye) {
        conctarget(xs, ys, 0, ye);
    }

    void conctarget(int xs, double ys, double dys, double ye) {
        float[] gr = gradient(gm(carboncycle.class).co2atppm, sy - 1); //returns y0, dy0, d2y0
        //make the pade curve
        if (curveopt.chosen==curvetype.padequartic) mathcurve.padequartic(target, sy - 1, xs, gr[0], gr[1], gr[2], ys, dys);
        if (curveopt.chosen==curvetype.padequintic) mathcurve.padequintic(target, sy - 1, xs, gr[0], gr[1], gr[2], ys, dys, 0); //final 0 is d2ys
        //after CO2 stabilisation point
        if (xs < gey) {
            try {
                if (ys == ye || (gey - xs < 20)) for (int x = xs; x <= gey; x++) target[x - gsy] = (float) ye; //flat
                else mathcurve.flatquadratic(target, xs, gey, ys, dys, ye);
            } catch (Exception e) {
                deb("conctarget last-part problem: xs="+xs); e.printStackTrace();
            }
        }
    } //end conctarget

    /* variants :
    growth 2000 = (7.77-4.20)*.471 =1.68 (emit-sink)*(ppmpgtc) ,	growth 1990 = (7.75-3.84)*.471 =1.84 ,	so d2y0 =  (1.68-1.84)/10 = -0.016
    could also fix dy0 =d2y0+1.65;
    double d2ys = padequartic( target, x0, xs, get(carboncycle.class).co2atppm[x0s],  dy0, d2y0, ys, dys);
    mathcurve.expdecay(target, xs, ys, dys, d2ys);
    mathcurve.simplepade (target, xs,  gey, ys, dys, d2ys, ye);
     */    //*****************************
    //STABILISE get(radfor.class) or TEMPERATURE by ITERATE CO2: guess CO2 stabilisation
    void setlevel() {
        if (ic == stabrf_allghgaero || ic == stabrf_allghg) {
            curvetocheck = ic == stabrf_allghgaero ? gm(radfor.class).anthrorf : gm(radfor.class).ghgrf;
            targetlevel = stabrflevel.getval();
            endtargetlevel = peaking.istrue() ? stabrfend.getval()  : targetlevel;
            setstabconcyear(stabyear.getval());
        }
        if (ic == stabtemp) {
            curvetocheck = gm(glotemp.class).avchange;
            targetlevel = stabtemplevel.getval() + gm(glotemp.class).offset; //offset because model temperature relative to 1750 but parameter relative to baseyear!
            endtargetlevel = peaking.istrue() ? stabtempend.getval() + gm(glotemp.class).offset : targetlevel;
            setstabconcyear(stabyear.getval());
        }
        if (ic == stabsea) {
            curvetocheck = gm(sealevel.class).total;
            targetlevel = stabsealevel.getval();
            endtargetlevel = targetlevel;
            setstabconcyear(stabseayear.getval());
        }
    }

    void setstabconcyear(double stabconcyear) {
        scy = (int) stabconcyear;
    //fsy+(int)(1.0*(stabconcyear-fsy)); //note here can experiment setting co2 stab earlier than temp stab year
    }
    //**************************
    //INITIAL GUESS
    //initial guesses for CO2 concentration target curve for stabilise get(radfor.class) or temp iterations
    //if first step of iteration set co2stab, scy, co2end, co2slope (otherwise these reset in correct method called from postcalc)
    void initialguess() {
        if (ic == stabrf_allghgaero || ic == stabrf_allghg) {
            guessconcfromrf(targetlevel * gm(radfor.class).totalrf.get(scy) / (ic == stabrf_allghgaero ? gm(radfor.class).anthrorf.get(scy) : gm(radfor.class).ghgrf.get(scy)));
        }
        if (ic == stabtemp) {
            guessconcfromrf(guessrffromtemp(targetlevel));
        }
        if (ic == stabsea) {
            guessconcfromrf(guessrffromtemp(guesstempfromsealevel(targetlevel)));
        }
    //		stabtemplevel.min=1.0-get(glotemp.class).offset; stabtemplevel.max=7.0-get(glotemp.class).offset; //set permitted range relative to 1750
    }

    double guesstempfromsealevel(double sealeveltarget) {
        return (gm(sealevel.class).total.get(scy) > 0 ? gm(glotemp.class).avchange.get(scy) * sealeveltarget / gm(sealevel.class).total.get(scy) : 2.0); //first guess - assume get(sealevel.class) proportional to temperature (in same year) - if get(sealevel.class) not yet calculated guess 2
    }

    double guessrffromtemp(double stabtemplevel) {	//called from precalc() -before steps
        //first guess - use ratio temp/get(radfor.class) if already calculated once, else use climate sensitivity assuming equilibrium
        return stabtemplevel * (gm(glotemp.class).avchange.get(scy) > 0 ? gm(radfor.class).totalrf.get(scy) / gm(glotemp.class).avchange.get(scy) : gm(udebclimod.class).val(udebclimod.cp.rfco2double) / gm(udebclimod.class).val(udebclimod.cp.climsens));
    }

    void guessconcfromrf(double guesstotrf) {
        double co2rf = guesstotrf * (gm(radfor.class).totalrf.get(scy) > 0 ? gm(carboncycle.class).co2rf.get(scy) / gm(radfor.class).totalrf.get(scy) : 0.85); //guess CO2 rf is same frac as current, or 85% if first time

        co2staborig = gm(carboncycle.class).atppmprein * Math.exp(Math.log(2.0) * co2rf / gm(udebclimod.class).val(udebclimod.cp.rfco2double)); //find corresponding co2 concentration

        //note: old corrections remembered from previous solution -faster when dragging control
        co2stab = co2staborig * oldstabcor;
        //if (co2stab<stabconclevel.min) {	co2stab=stabconclevel.min; debug("can't drop CO2 below "+stabconclevel.min); } 	//catch impossible target
        co2stabslopeorig = 0;
        co2stabslope = oldstabslopecor;
        co2endorig = co2stab;
        co2end = co2endorig * oldendcor;
    }
    //**************************
    //APPLY CORRECTION
    //for iteration: correct according to calculated rf or temp, called from postcalc()
    void applycorrection() {

        //correct level in stabilisation year
        stabcor = targetlevel / curvetocheck.get(scy);

        //gradient should be flat at stabilisation point (check twenty years gradient)
        stabslopecor = -0.05 * (curvetocheck.get(scy + 10) - curvetocheck.get(scy - 10)) / curvetocheck.get(scy - 1);

        //end level should be same as stab level: (check penultimate step - note use actual stab level rather than target level)
        endcor = (curvetocheck.get(scy) / curvetocheck.get(gey - 1)) * (endtargetlevel / targetlevel);

        //end gradient should be flat (check last twenty years gradient)
        //endslopecor = -0.05*(curvetocheck[(gey-gsy)-1]-curvetocheck[(gey-gsy)-21]) / curvetocheck[sys-1];


        co2stab = carboncycle.atppmprein + (co2stab - carboncycle.atppmprein) * stabcor;
        if (co2stab < 300) co2stab = 300;
        if (stit <= 6 || stabslopecor / oldstabslopecor < 1.0) co2stabslope += stabslopecor * (co2stab - carboncycle.atppmprein); //only use if iteration 0 - 6 or correction is getting smaller - avoid instability
        co2end = carboncycle.atppmprein + (co2end - carboncycle.atppmprein) * endcor;
        if (co2end < 300) co2end = 300;
        //if (stit<=4 || endslopecor/oldendslopecor <1.0)  co2endslope+=endslopecor * (co2stab-carboncycle.atppmprein);

        //stop iteration if impossible targets
        boolean impossible = (Double.isInfinite(co2stab) || Double.isInfinite(co2end) || Double.isNaN(co2stab) || Double.isNaN(co2end) || co2stab < 300 || co2end < 300);

        //P3 note oct08: added precision factor of ten - check OK and  still fast enough...
        goodenough = (impossible || ((stabcor < 1.001) && (stabcor > 0.999) && (stabslopecor < 0.001) && (stabslopecor > -0.001) && (endcor < 1.001) && (endcor > 0.999) //&& (endslopecor<0.05) && (endslopecor>-0.05)
                ));
        /*
        (stabcor<1.005) && (stabcor>0.995) 	&& (stabslopecor<0.02) && (stabslopecor>-0.02)
        && (endcor<1.005) && (endcor>0.995) // && (endslopecor<0.01) && (endslopecor>-0.01)
         */

        stit++;
        savecorrect();

    } //end applycorrection

    void savecorrect() {	//saving oldcorrect  makes iteration faster if dragging arrow
        oldstabcor = co2stab / co2staborig;
        oldstabslopecor = (co2stabslope - co2stabslopeorig) / (co2stab - carboncycle.atppmprein);
        //oldmidcor=co2mid/co2midorig;
        oldendcor = co2end / co2endorig; //oldendslopecor=(co2endslope-co2endslopeorig) /   (co2stab-carboncycle.atppmprein);
    }

    void resetcorrect() {
        oldstabcor = 1.0;
        oldstabslopecor = 0; /*oldmidcor=1.0;*/ oldendcor = 1.0; /* oldendslopecor=0; */ }

    void resetguess() {
        resetcorrect();
        goodenough = false;
    } //see climate module

    public void run() {
        running = true;
        log("Stabilisation iteratation starting");
        //Thread.currentThread().setPriority(Thread.MIN_PRIORITY);
        Thread.currentThread().yield();
        loop.waitUntilLoopDone();

        while (!goodenough && stit < 20) {
            changed = true;
            loop.gonow(false);
        }
        log("Stabilisation iteratated " + stit + " times");
        //log(	"co2stabslope "+co2stabslope+" ");

        stit = 0;
        running = false;
    /*if (objective.chosen=="stabsea") resetcorrect(); else */
    }
    /*
    info for debugging - put after calc goodenough
    + " STAB= " + aa(co2stab) 	+" corr= " + aa(stabcor) + " slope= " + aa(co2stabslope) +" corr= "+aa(stabslopecor) 	//+ "MID= " + aa(co2mid) +" corr= "+ aa(midcor)
    + " END= " + aa(co2end) +" corr= "+ aa(endcor) 	+ " slope= " + aa(co2endslope) +" corr= "+aa(endslopecor)
    +" ok? "+goodenough+"\n" );
    double aa(double d) {	return ((int) (d*10000.0))/10000.0; }
     */
} //end stabilisation class.







