/*
Other (non co2) greenhouse gases and aerosols,
CH4, N2O, O3, Nox, SOx, VOC, CO, SF6, HFCs, PFCs, CFCs, HCFCs
SRES Emissions which may be scaled to CO2 reduction)
Atmospheric chemistry including Ozone/OH
Radiative forcing for these gases
Applies formula from Bern-CC / IPCC WG1ch6
 
//Developed February 2002, Bern
 
 */
package jcm.mod.ogas;
import jcm.core.*;
import jcm.core.par.param;
import jcm.core.cur.curve;
import jcm.core.cur.curveset;
import jcm.core.ob.module;
import static jcm.gui.gen.colfont.*;
import static jcm.core.complexity.*;
import jcm.mod.cli.radfor;

public class atchem extends module  {
    
    
    //*******************************************************
    //INTERACTIONS
    
    
    
    //PARAMS
    public param
            oheffect=new param("OHef", true, expert), 	 //note this also controls N2O life
            taro3=new param("TARO3", false, expert);
    
    //PLOTARRAYS
    public curve
            ch4conc=new curve( "ch4", green ),
            n2oconc=new curve( "n2o" , brown ),
            tropo3du=new curve(  "tropo3du", grey ),
            reloh =new curve( "reloh" ,  dkgrey ); //OH relative to 2000;
    public static curve
            ch4concmeas=new curve( "ch4-data", dkgreen, expert, 1750, 2009 ),
            n2oconcmeas=new curve( "n2o-data" , dkbrown, expert, 1750, 2009);
    
    public curve
            ch4rf= new curve( "ch4", green  ),
            n2orf= new curve("n2o" , brown),
            tropo3rf= new curve( "tropo3", grey),
            strath2orf= new curve( "strath2o" , yellowgreen ,  expert),
            othgasrf =new curve( "otherghg",  dkgreen, curve.Type.total );
    
    public curveset conc=new curveset( ch4conc, ch4concmeas, n2oconc, n2oconcmeas, tropo3du, reloh, this.name+"&conc", "ppb");
    public curveset rf=new curveset(	ch4rf, n2orf, tropo3rf, strath2orf, othgasrf,  this.name+"&rf", "w&per&m2");
    
    
    
    //*******************************************************
    //LOOP CALCS
    
    //before step-loop
    public void initsetup() {
        super.initsetup();
        initialdata();
        rf.addcurve(  gm(fgas.class).fgasrf );
        follows(othgasemit.class);
        setaffectedby(gm(radfor.class).applyefficacy);
    }
    
    public void calcstep(){
        calcreloh();
        //used to fix historical concentrations of CH4, N2O, now vary them to show effect of lifetime, and for consistency with responsibility calculations
        if (year>gsy) {	ch4conc(); n2oconc(); }
        if (year>=2000) tropo3conc();
        oghgrf();
        calctot();
    } //end calcstep
    
    
    //*******************************************************
    void initialdata() {
/*
 CH4 and N2O concentration data
  note quite high precision needed if using marginal d(rf)/d(conc) in responsibility
CH4 orig from 1750 to 2000 copied here from old JCM histdata.java, need to trace original source
from 2001-2006 added from Mauna Loa (monthly flask data averaged over year, from WGGDC) - subtract 3 which is the average difference 1992-2000
 update 2009 http://agage.eas.gatech.edu/images/Data_figures/gcmd_month/ch4_monS5.pdf
 Note difference between stations about 150ppb
 Samoa + Tasmania (SH clean air) most similar - avg 1735 around 2000-2007, up last two years to 1755 in 2009,  cf Mauna Loa 200-2006 about 1765, so add three points 1765,  1775, 1785



 N2O orig from 1750 to 1998 copied here from old JCM histdata.java, need to trace original source
data from Mauna Loa show almost linear trend from 1993-2006, 315.0 in 1998 and 321.5 in 2006, - add the 0.8/yr trend to end of old data
update 2009: http://agage.eas.gatech.edu/images/Data_figures/gcmd_month/n2o_monS5.pdf range of stations shows about 314 in 1998, 320 in 2006, 322.4 in 2009, linear trend continues
 *
 tropospheric ozone data is in dobson units, from 1850-2000 dobson units (assume constant at 25 before 1850)
 note: mauna loa data also available for O3 to 2004 and CO to 2006 (both in ppb) , but variability much greater than trend
 regional data might be more useful for these gases
 
 */
        
        float[] ch4concdata =	{732.46f, 733.18f, 733.89f, 734.59f, 735.29f, 735.97f, 736.65f, 737.31f, 737.96f, 738.60f, 739.23f, 739.85f, 740.46f, 741.05f, 741.64f, 742.22f, 742.78f, 743.34f, 743.90f, 744.44f, 744.97f, 745.50f, 746.02f, 746.54f, 747.05f, 747.56f, 748.06f, 748.56f, 749.05f, 749.54f, 750.03f, 750.51f, 751f, 751.49f, 751.99f, 752.49f, 753f, 753.51f, 754.03f, 754.57f, 755.11f, 755.67f, 756.25f, 756.84f, 757.44f, 758.07f, 758.71f, 759.38f, 760.07f, 760.78f, 761.51f, 762.28f, 763.08f, 763.90f, 764.76f, 765.65f, 766.57f, 767.51f, 768.48f, 769.48f, 770.49f, 771.53f, 772.58f, 773.65f, 774.74f, 775.84f, 776.95f, 778.08f, 779.21f, 780.36f, 781.51f, 782.68f, 783.85f, 785.02f, 786.20f, 787.38f, 788.57f, 789.75f, 790.93f, 792.12f, 793.29f, 794.47f, 795.63f, 796.79f, 797.94f, 799.07f, 800.20f, 801.31f, 802.40f, 803.47f, 804.53f, 805.56f, 806.59f, 807.59f, 808.59f, 809.59f, 810.58f, 811.58f, 812.59f, 813.60f, 814.63f, 815.68f, 816.74f, 817.82f, 818.93f, 820.06f, 821.20f, 822.36f, 823.54f, 824.74f, 825.95f, 827.18f, 828.43f, 829.70f, 830.98f, 832.27f, 833.59f, 834.91f, 836.26f, 837.62f, 839f, 840.39f, 841.80f, 843.24f, 844.69f, 846.17f, 847.66f, 849.19f, 850.73f, 852.31f, 853.91f, 855.55f, 857.21f, 858.92f, 860.66f, 862.44f, 864.27f, 866.14f, 868.06f, 870.04f, 872.07f, 874.15f, 876.30f, 878.51f, 880.78f, 883.12f, 885.53f, 888f, 890.55f, 893.16f, 895.85f, 898.61f, 901.44f, 904.35f, 907.33f, 910.39f, 913.53f, 916.75f, 920.04f, 923.42f, 926.87f, 930.41f, 934.03f, 937.74f, 941.53f, 945.41f, 949.38f, 953.43f, 957.58f, 961.82f, 966.15f, 970.59f, 975.12f, 979.75f, 984.49f, 989.34f, 994.30f, 999.39f, 1004.59f, 1009.93f, 1015.41f, 1021.03f, 1026.81f, 1032.75f, 1038.85f, 1045.13f, 1051.59f, 1058.24f, 1065.08f, 1072.13f, 1079.39f, 1086.87f, 1094.58f, 1102.53f, 1110.72f, 1119.17f, 1127.88f, 1136.86f, 1146.11f, 1155.65f, 1165.48f, 1175.60f, 1186.03f, 1196.76f, 1207.79f, 1219.13f, 1230.77f, 1242.71f, 1254.94f, 1267.47f, 1280.29f, 1293.39f, 1306.75f, 1320.37f, 1334.24f, 1348.35f, 1362.70f, 1377.27f, 1392.05f, 1407.03f, 1422.20f, 1437.53f, 1452.99f, 1468.57f, 1484.22f, 1499.92f, 1515.60f, 1531.23f, 1546.76f, 1562.13f, 1577.29f, 1592.20f, 1606.79f, 1621.02f, 1634.83f, 1648.17f, 1661.02f, 1673.37f, 1685.24f, 1696.69f, 1707.78f, 1718.63f, 1729.34f, 1723.5f, 1728.7f, 1733.9f, 1739.1f, 1744.3f, 1749.6f, 1754.8f, 1760.0f};
        double[] ch4ml9206= {1721.7, 1726.98, 1730.23, 1738.98, 1740.9, 1748.36, 1757.76, 1761.54, 1763.64, 1766.38, 1762.24, 1772.29, 1770.49, 1765.35, 1764.1, 1765, 1775, 1785}; //last three estimates -see notes
        
        float[] n2oconcdata=	{271.68f, 271.70f, 271.72f, 271.74f, 271.77f, 271.79f, 271.81f, 271.84f, 271.86f, 271.88f, 271.91f, 271.93f, 271.95f, 271.98f, 272f, 272.02f, 272.05f, 272.07f, 272.09f, 272.12f, 272.14f, 272.17f, 272.19f, 272.21f, 272.24f, 272.26f, 272.29f, 272.31f, 272.34f, 272.36f, 272.38f, 272.41f, 272.44f, 272.46f, 272.49f, 272.51f, 272.54f, 272.56f, 272.59f, 272.62f, 272.64f, 272.67f, 272.70f, 272.72f, 272.75f, 272.78f, 272.81f, 272.83f, 272.86f, 272.89f, 272.92f, 272.95f, 272.98f, 273.01f, 273.04f, 273.07f, 273.10f, 273.13f, 273.16f, 273.19f, 273.22f, 273.26f, 273.29f, 273.32f, 273.36f, 273.39f, 273.43f, 273.46f, 273.50f, 273.53f, 273.57f, 273.61f, 273.64f, 273.68f, 273.72f, 273.76f, 273.80f, 273.84f, 273.88f, 273.92f, 273.96f, 274.01f, 274.05f, 274.09f, 274.14f, 274.19f, 274.23f, 274.28f, 274.33f, 274.37f, 274.42f, 274.47f, 274.52f, 274.57f, 274.63f, 274.68f, 274.73f, 274.79f, 274.84f, 274.90f, 274.96f, 275.01f, 275.07f, 275.13f, 275.19f, 275.25f, 275.32f, 275.38f, 275.44f, 275.51f, 275.57f, 275.64f, 275.71f, 275.78f, 275.85f, 275.92f, 275.99f, 276.07f, 276.14f, 276.22f, 276.29f, 276.37f, 276.45f, 276.53f, 276.61f, 276.69f, 276.78f, 276.86f, 276.95f, 277.03f, 277.12f, 277.21f, 277.30f, 277.39f, 277.49f, 277.58f, 277.68f, 277.77f, 277.87f, 277.97f, 278.07f, 278.18f, 278.28f, 278.38f, 278.49f, 278.60f, 278.71f, 278.82f, 278.93f, 279.05f, 279.16f, 279.28f, 279.40f, 279.52f, 279.64f, 279.76f, 279.88f, 280.01f, 280.14f, 280.27f, 280.39f, 280.53f, 280.66f, 280.79f, 280.93f, 281.07f, 281.20f, 281.34f, 281.48f, 281.63f, 281.77f, 281.91f, 282.05f, 282.20f, 282.34f, 282.48f, 282.63f, 282.77f, 282.92f, 283.07f, 283.21f, 283.36f, 283.51f, 283.66f, 283.81f, 283.97f, 284.13f, 284.28f, 284.45f, 284.61f, 284.78f, 284.95f, 285.13f, 285.32f, 285.51f, 285.70f, 285.90f, 286.11f, 286.33f, 286.55f, 286.78f, 287.02f, 287.28f, 287.54f, 287.81f, 288.09f, 288.38f, 288.69f, 289.01f, 289.34f, 289.68f, 290.04f, 290.42f, 290.80f, 291.21f, 291.63f, 292.07f, 292.52f, 292.99f, 293.48f, 293.99f, 294.52f, 295.06f, 295.63f, 296.21f, 296.82f, 297.45f, 298.10f, 298.77f, 299.47f, 300.18f, 300.92f, 301.68f, 302.45f, 303.24f, 304.03f, 304.82f, 305.61f, 306.39f, 307.16f, 307.93f, 308.69f, 309.44f, 310.19f, 310.92f, 311.65f, 312.38f, 313.11f, 313.83f};
        double[] tropo3dudata=      {25.0, 25.0, 25.0, 25.1, 25.1, 25.1, 25.1, 25.1, 25.2, 25.2, 25.2, 25.2, 25.3, 25.3, 25.3, 25.3, 25.3, 25.4, 25.4, 25.4, 25.4, 25.5, 25.5, 25.5, 25.5, 25.5, 25.5, 25.6, 25.6, 25.6, 25.6, 25.7, 25.7, 25.7, 25.7, 25.7, 25.8, 25.8, 25.8, 25.8, 25.9, 25.9, 25.9, 25.9, 25.9, 26.0, 26.0, 26.0, 26.0, 26.0, 26.1, 26.1, 26.1, 26.2, 26.2, 26.2, 26.3, 26.3, 26.3, 26.4, 26.4, 26.4, 26.5, 26.5, 26.5, 26.6, 26.6, 26.6, 26.7, 26.7, 26.7, 26.8, 26.8, 26.8, 26.9, 26.9, 26.9, 27.0, 27.0, 27.0, 27.1, 27.1, 27.1, 27.2, 27.2, 27.2, 27.3, 27.3, 27.3, 27.4, 27.4, 27.4, 27.5, 27.5, 27.5, 27.6, 27.6, 27.6, 27.7, 27.7, 27.7, 27.8, 27.9, 28.0, 28.1, 28.2, 28.3, 28.4, 28.5, 28.6, 28.7, 28.9, 29.0, 29.1, 29.2, 29.4, 29.5, 29.6, 29.7, 29.9, 30.0, 30.1, 30.2, 30.3, 30.5, 30.6, 30.7, 30.8, 30.9, 31.0, 31.2, 31.3, 31.5, 31.6, 31.7, 31.9, 32.0, 32.2, 32.3, 32.5, 32.6, 32.7, 32.9, 33.0, 33.2, 33.3, 33.4, 33.6, 33.7, 33.9, 34.0};
        
        for (int j=1750; j<2010; j++) ch4concmeas.set(j, j<=2000 ? ch4concdata[j-1750] : (float)(ch4ml9206[j-1992]-3.0));
        for (int j=1750; j<2010; j++) n2oconcmeas.set(j, j<=1998 ? n2oconcdata[j-1750] :  n2oconcmeas.get(j-1)+0.8f);
        for (int j=1750; j<=2000; j++) tropo3du.set(j, j<1850 ? 25f : (float)tropo3dudata[j-1850]);
        
        //P3 CHECK logic here: due to feedbacks in atmospheric chemistry, have to set the model conc too, later it will change with emissions
        for (int j=1750; j<2010; j++) ch4conc.set(j, ch4concmeas.get(j));
        for (int j=1750; j<2010; j++) n2oconc.set(j, n2oconcmeas.get(j));
            
        histn2o();
        oghgrf0();
    }
    
    
    //*************** CHEMISTRY ****************************************

    //CH4, CO, VOC, NOX, => OH, tropO3
    float tropo3prein=25;
    
    //calculate OH relative to 2000
    void calcreloh() {
        if (oheffect.istrue()) {
            //note ch4 <=> oh, calc oh first using previous year's ch4 projected
            float projectch4conc= year>gsy+1 ? 2*ch4conc.get(year-1)-ch4conc.get(year-2) : ch4prein;
            othgasemit othgasemit=gm(othgasemit.class);
            reloh.set(year,  (float)Math.exp(
                    -0.32 * Math.log(projectch4conc/ch4conc.get(2000))
                    +0.0042 *(othgasemit.noxemit.get(year)-othgasemit.noxemit.get(2000))
                    -0.000105 *(othgasemit.coemit.get(year)-othgasemit.coemit.get(2000))
                    -0.000315 *(othgasemit.vocemit.get(year)-othgasemit.vocemit.get(2000))
                    ));
        } else reloh.set(year, 1f);
    }
    
    //TROP O3 calculate tropospheric ozone (dobson units) - only for future
    void tropo3conc() {
        othgasemit othgasemit=gm(othgasemit.class);
        tropo3du.set(year, tropo3du.get(2000)+(float)(
                taro3.istrue() ? ( 	//coefficients used in TAR -too sensitive, can go below preindustrial level (hence rf<0) if emissions drop a lot
                6.7 * Math.log(ch4conc.get(year)/ch4conc.get(2000))
                +0.17 *(othgasemit.noxemit.get(year)-othgasemit.noxemit.get(2000))
                +0.0014 *(othgasemit.coemit.get(year)-othgasemit.coemit.get(2000))
                +0.0042 *(othgasemit.vocemit.get(year)-othgasemit.vocemit.get(2000))
                ) : ( 	//later corrected coefficients from Prather -see also Joos paper
                5.0 * Math.log(ch4conc.get(year)/ch4conc.get(2000))
                +0.125 *(othgasemit.noxemit.get(year)-othgasemit.noxemit.get(2000))
                +0.0011 *(othgasemit.coemit.get(year)-othgasemit.coemit.get(2000))
                +0.0033 *(othgasemit.vocemit.get(year)-othgasemit.vocemit.get(2000))
                )
                ));
        if (tropo3du.get(year)<0) tropo3du.set(year, 0);
    }
    
    //***************** N2O *******************
    
    public float  ppbpmtn_old=0.20789f, n2oprein=272f, ppbpmtn2o=ppbpmtn_old*28f/44f; //last needed since changed emissions curve units from N to N2O
    //n2o lifetime: affected by n2o concentration or emissions (as implied in IPCC / Joos / Prather ?)
    //float n2olife() {return 120f*(float)Math.pow((n2oconc.get(year))/(n2oconc.get(2000)),-0.055);}
    public float n2olife(int y) {
        othgasemit othgasemit=gm(othgasemit.class);
        return 120f * (oheffect.istrue() ? (float)Math.pow((n2oemitnat()+othgasemit.n2oemit.get(y))/(n2oemitnat()+othgasemit.n2oemit.get(2000)),-0.055) : 1f) ;
    } //note constant if OHeffect false - added for ACCC default
    
    public float n2oemitnat() {	return (44f/28f) * (oheffect.istrue() ? 10.6f : 10.9f ) ; /*9.5f*/ }
                        /*
                        added 44/28 when changed units from Mt N to Mt N2O
                        n2o emit in 2000 is about 7.1,
                        so if no anthrop emissions, to balance n2oemitnat and n2olife (variable), need n2oemitnat()=10.6
                        but original figure from TAR methods was 9.5
                        if no feedback on lifetime, need to start with 10.9
                         */
    
    //calculate future n2o conc from emit
    void n2oconc() {
        n2oconc.set(year, n2oconc.get(year-1)*(1f-1f/n2olife(year-1))+ppbpmtn2o*(n2oemitnat()+gm(othgasemit.class).n2oemit.get(year)));
    }
    
    //calculate historical n2o emit from measured concentration - but note: this needs emissions to get n2olife! => set it going with emit in 2007
    void histn2o() {
        gm(othgasemit.class).n2oemit.set(2006, 7f); //just for n2Olife, will be overwritten later
        for (int y=2006; y>gsy; y--) gm(othgasemit.class).n2oemit.set(y-1,  ( (n2oconcmeas.get(y)-n2oconcmeas.get(y-1)*(1f-1f/n2olife(y))) /ppbpmtn2o)-n2oemitnat());
    }
    
    //****************** CH4 *****************
    //puzzle regarding ppbpmtch4: derived from ppmpmtc should be 0.35325, joos paper has ppbpmtch4=0.3603, tar says 1/2.78 =0.35971
    public static float ppbpmtch4=0.36f, ch4prein=742.2f;
    
    public float ch4emitnat() {
        if (oheffect.istrue()) return 301f; //original TAR
        else return ch4prein  / (ch4life(gsy) *ppbpmtch4); //consistent with initial lifetime and concentration:
    }
    
    public float ch4life(int y){	return 1f/((reloh.get(y)/9.58f)+(1f/68.2f)); }
    
    void ch4conc() {	ch4conc.set(year,    ch4conc.get(year-1)*(1f-1f/ch4life(year))+ppbpmtch4*(ch4emitnat()+gm(othgasemit.class).ch4emit.get(year))	    ); }
    //note, if no OH effect, would need to use a different natural CH4 flux to start in steady state
    
    //****************** OGHG FORCING *************************************
    
    
    float overlap0, ch4rf0, n2orf0;
    
    void oghgrf0() {
        overlap0=overlap(ch4conc.get(gsy), n2oconc.get(gsy));
        ch4rf0=(float)(0.036 * Math.pow(ch4conc.get(gsy),0.5));
        n2orf0=(float)(0.12 * Math.pow(n2oconc.get(gsy),0.5));
    }
    
    void oghgrf() {
        ch4rf.set(year, (float)((0.036*Math.pow(ch4conc.get(year),0.5) -ch4rf0) -(overlap(ch4conc.get(year),n2oconc.get(gsy))-overlap0)) *  gm(radfor.class).getefficacy(gas.ch4));
        n2orf.set(year, (float)((0.12*Math.pow(n2oconc.get(year),0.5) -n2orf0) -(overlap(ch4conc.get(gsy),n2oconc.get(year))-overlap0)) *  gm(radfor.class).getefficacy(gas.n2o));
        tropo3rf.set(year, (float) (0.042*(tropo3du.get(year)-tropo3prein)) *  gm(radfor.class).getefficacy(gas.o3trop));
        strath2orf.set(year, -0.05f*ch4rf.get(year));
    }
    
    //overlap between ch4 and n2o rf
    float overlap(float ch4, float n2o) {
        return (float)(0.47* Math.log(
                1.0 + 2.01e-5 * Math.pow(ch4*n2o,0.75)
                + 5.31e-15 * ch4 * Math.pow(ch4*n2o,1.52)
                ));
    }
    
    void calctot() {
        othgasrf.set(year,  ch4rf.get(year)+n2orf.get(year) +tropo3rf.get(year)+strath2orf.get(year));
    }
    
} //end class




